@@ -47,20 +47,6 @@ static void     nemo_action_set_property  (GObject                    *object,
 static void     nemo_action_constructed (GObject *object);
 static void     nemo_action_finalize (GObject *gobject);
 
-static SelectionType nemo_action_get_selection_type   (NemoAction *action);
-static void          nemo_action_set_extensions       (NemoAction *action, gchar **extensions);
-static gchar       **nemo_action_get_extension_list   (NemoAction *action);
-static void          nemo_action_set_mimetypes        (NemoAction *action, gchar **mimetypes);
-static gchar       **nemo_action_get_mimetypes_list   (NemoAction *action);
-static void          nemo_action_set_key_file_path    (NemoAction *action, const gchar *path);
-static void          nemo_action_set_exec             (NemoAction *action, const gchar *exec);
-static void          nemo_action_set_parent_dir       (NemoAction *action, const gchar *parent_dir);
-static void          nemo_action_set_separator        (NemoAction *action, const gchar *separator);
-static void          nemo_action_set_conditions       (NemoAction *action, gchar **conditions);
-static gchar       **nemo_action_get_conditions       (NemoAction *action);
-static void          nemo_action_set_orig_label       (NemoAction *action, const gchar *orig_label);
-static void          nemo_action_set_orig_tt          (NemoAction *action, const gchar *orig_tt);
-
 static gchar   *find_token_type (const gchar *str, TokenType *token_type);
 
 static gpointer parent_class;
@@ -81,7 +67,9 @@ enum
   PROP_QUOTE_TYPE,
   PROP_ESCAPE_SPACE,
   PROP_RUN_IN_TERMINAL,
-  PROP_CONDITIONS
+  PROP_CONDITIONS,
+  PROP_URI_SCHEME,
+  PROP_LAST
 };
 
 enum {
@@ -155,6 +143,7 @@ nemo_action_init (NemoAction *action)
     action->escape_space = FALSE;
     action->show_in_blank_desktop = FALSE;
     action->run_in_terminal = FALSE;
+    action->uri_scheme = NULL;
 
     action->constructing = TRUE;
 }
@@ -291,6 +280,15 @@ nemo_action_class_init (NemoActionClass *klass)
                                                            G_PARAM_READWRITE)
                                  );
 
+    g_object_class_install_property (object_class,
+                                     PROP_URI_SCHEME,
+                                     g_param_spec_string ("uri-scheme",
+                                                          "Limit selection by uri scheme (like file, sftp, etc...)",
+                                                          "Limit selection by uri scheme (like file, sftp, etc...)",
+                                                          NULL,
+                                                          G_PARAM_READWRITE)
+                                     );
+
     signals[CONDITION_CHANGED] = g_signal_new ("condition-changed",
                                                G_TYPE_FROM_CLASS (object_class),
                                                G_SIGNAL_RUN_LAST,
@@ -699,6 +697,11 @@ nemo_action_constructed (GObject *object)
                                               KEY_SEPARATOR,
                                               NULL);
 
+    gchar *uri_scheme = g_key_file_get_string (key_file,
+                                               ACTION_FILE_GROUP,
+                                               KEY_URI_SCHEME,
+                                               NULL);
+
     gchar *quote_type_string = g_key_file_get_string (key_file,
                                                       ACTION_FILE_GROUP,
                                                       KEY_QUOTE_TYPE,
@@ -843,6 +846,7 @@ nemo_action_constructed (GObject *object)
                    "conditions", conditions,
                    "escape-space", escape_space,
                    "run-in-terminal", run_in_terminal,
+                   "uri-scheme", uri_scheme,
                     NULL);
 
     action->constructing = FALSE;
@@ -860,6 +864,7 @@ nemo_action_constructed (GObject *object)
     g_free (parent_dir);
     g_free (quote_type_string);
     g_free (separator);
+    g_free (uri_scheme);
     g_strfreev (ext);
     g_strfreev (mimes);
     g_strfreev (conditions);
@@ -995,6 +1000,7 @@ nemo_action_finalize (GObject *object)
     g_free (action->orig_label);
     g_free (action->orig_tt);
     g_free (action->separator);
+    g_free (action->uri_scheme);
 
     if (action->dbus) {
         g_list_free_full (action->dbus, (GDestroyNotify) dbus_condition_free);
@@ -1033,40 +1039,40 @@ nemo_action_set_property (GObject         *object,
   switch (prop_id)
     {
     case PROP_KEY_FILE_PATH:
-      nemo_action_set_key_file_path (action, g_value_get_string (value));
+      action->key_file_path = g_strdup (g_value_get_string (value));
       break;
     case PROP_SELECTION_TYPE:
       action->selection_type = g_value_get_int (value);
       break;
     case PROP_EXTENSIONS:
-      nemo_action_set_extensions (action, g_value_get_pointer (value));
+      action->extensions = g_strdupv (g_value_get_pointer (value));
       break;
     case PROP_MIMES:
-      nemo_action_set_mimetypes (action, g_value_get_pointer (value));
+      action->mimetypes = g_strdupv (g_value_get_pointer (value));
       break;
     case PROP_EXEC:
-      nemo_action_set_exec (action, g_value_get_string (value));
+      action->exec = g_strdup (g_value_get_string (value));
       break;
     case PROP_PARENT_DIR:
-      nemo_action_set_parent_dir (action, g_value_get_string (value));
+      action->parent_dir = g_strdup (g_value_get_string (value));
       break;
     case PROP_USE_PARENT_DIR:
       action->use_parent_dir = g_value_get_boolean (value);
       break;
     case PROP_ORIG_LABEL:
-      nemo_action_set_orig_label (action, g_value_get_string (value));
+      action->orig_label = g_strdup (g_value_get_string (value));
       break;
     case PROP_ORIG_TT:
-      nemo_action_set_orig_tt (action, g_value_get_string (value));
+      action->orig_tt = g_strdup (g_value_get_string (value));
       break;
     case PROP_QUOTE_TYPE:
       action->quote_type = g_value_get_int (value);
       break;
     case PROP_SEPARATOR:
-      nemo_action_set_separator (action, g_value_get_string (value));
+      action->separator = g_strdup (g_value_get_string (value));
       break;
     case PROP_CONDITIONS:
-      nemo_action_set_conditions (action, g_value_get_pointer (value));
+      action->conditions = g_strdupv (g_value_get_pointer (value));
       break;
     case PROP_ESCAPE_SPACE:
       action->escape_space = g_value_get_boolean (value);
@@ -1074,6 +1080,9 @@ nemo_action_set_property (GObject         *object,
     case PROP_RUN_IN_TERMINAL:
       action->run_in_terminal = g_value_get_boolean (value);
       break;
+    case PROP_URI_SCHEME:
+      action->uri_scheme = g_strdup (g_value_get_string (value));
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1134,6 +1143,9 @@ nemo_action_get_property (GObject    *object,
     case PROP_RUN_IN_TERMINAL:
       g_value_set_boolean (value, action->run_in_terminal);
       break;
+    case PROP_URI_SCHEME:
+      g_value_set_string (value, action->uri_scheme);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
@@ -1157,10 +1169,14 @@ find_token_type (const gchar *str, TokenType *token_type)
             *token_type = TOKEN_URI_LIST;
             return ptr;
         }
-        if (g_str_has_prefix (ptr, TOKEN_EXEC_PARENT)) {
+        if (g_str_has_prefix (ptr, TOKEN_EXEC_LOCATION_PATH)) {
             *token_type = TOKEN_PARENT_PATH;
             return ptr;
         }
+        if (g_str_has_prefix (ptr, TOKEN_EXEC_LOCATION_URI)) {
+            *token_type = TOKEN_PARENT_URI;
+            return ptr;
+        }
         if (g_str_has_prefix (ptr, TOKEN_EXEC_FILE_NAME)) {
             *token_type = TOKEN_FILE_DISPLAY_NAME;
             return ptr;
@@ -1197,26 +1213,21 @@ find_token_type (const gchar *str, TokenType *token_type)
 static gchar *
 get_path (NemoAction *action, NemoFile *file)
 {
-    gchar *ret, *quote_escaped, *orig;
+    gchar *ret, *orig;
 
     orig = nemo_file_get_path (file);
 
-    if (action->quote_type != QUOTE_TYPE_DOUBLE && action->quote_type != QUOTE_TYPE_SINGLE)
-        quote_escaped = eel_str_escape_quotes (orig);
-    else
-        quote_escaped = orig;
-
-    if (action->escape_space) {
-        ret = eel_str_escape_spaces (quote_escaped);
+    if (action->quote_type == QUOTE_TYPE_DOUBLE) {
+        ret = eel_str_escape_double_quoted_content (orig);
+    } else if (action->quote_type == QUOTE_TYPE_SINGLE) {
+        // Replace literal ' with a close ', a \', and an open '
+        ret = eel_str_replace_substring (orig, "'", "'\\''");
     } else {
-        ret = g_strdup (quote_escaped);
+        ret = eel_str_escape_shell_characters (orig);
     }
 
     g_free (orig);
 
-    if (quote_escaped != orig)
-        g_free (quote_escaped);
-
     return ret;
 }
 
@@ -1270,17 +1281,25 @@ static gchar *
 get_device_path (NemoAction *action, NemoFile *file)
 {
     GMount *mount = nemo_file_get_mount (file);
-    GVolume *volume = g_mount_get_volume (mount);
-    gchar *ret = NULL;
 
-    if (action->escape_space) {
-        gchar *id = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
-        ret = eel_str_escape_spaces (id);
-        g_free (id);
+    g_return_val_if_fail (mount != NULL, NULL);
+
+    GVolume *volume = g_mount_get_volume (mount);
+    gchar *ret, *id;
+
+    id = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
+
+    if (action->quote_type == QUOTE_TYPE_DOUBLE) {
+        ret = eel_str_escape_double_quoted_content (id);
+    } else if (action->quote_type == QUOTE_TYPE_SINGLE) {
+        // Replace literal ' with a close ', a \', and an open '
+        ret = eel_str_replace_substring (id, "'", "'\\''");
     } else {
-        ret = g_volume_get_identifier (volume, G_VOLUME_IDENTIFIER_KIND_UNIX_DEVICE);
+        ret = eel_str_escape_shell_characters (id);
     }
 
+    g_free (id);
+
     g_object_unref (mount);
     g_object_unref (volume);
 
@@ -1336,7 +1355,7 @@ get_insertion_string (NemoAction *action,
                     first = FALSE;
                 }
             } else {
-                goto default_parent_path;
+                goto default_parent_uri;
             }
             break;
         case TOKEN_PARENT_PATH:
@@ -1357,6 +1376,33 @@ default_parent_path:
             str = insert_quote (action, str);
             g_free (path);
             break;
+        case TOKEN_PARENT_URI:
+            ;
+default_parent_uri:
+            ;
+            gchar *uri;
+            gchar *name = nemo_file_get_display_name (parent);
+            if (g_strcmp0 (name, "x-nemo-desktop") == 0) {
+                gchar *real_desktop_path = nemo_get_desktop_directory ();
+                if (real_desktop_path) {
+                    GFile *file;
+                    file = g_file_new_for_path (real_desktop_path);
+                    uri = g_file_get_uri (file);
+                    g_object_unref (file);
+                    g_free (real_desktop_path);
+                } else {
+                    uri = NULL;
+                }
+            } else {
+                uri = nemo_file_get_uri (parent);
+            }
+
+            str = insert_quote (action, str);
+            str = score_append (action, str, uri);
+            str = insert_quote (action, str);
+            g_free (name);
+            g_free (uri);
+            break;
         case TOKEN_FILE_DISPLAY_NAME:
             if (g_list_length (selection) > 0) {
                 gchar *file_display_name = nemo_file_get_display_name (NEMO_FILE (selection->data));
@@ -1503,119 +1549,6 @@ nemo_action_activate (NemoAction *action,
     g_string_free (exec, TRUE);
 }
 
-static SelectionType
-nemo_action_get_selection_type (NemoAction *action)
-{
-    return action->selection_type;
-}
-
-static void
-nemo_action_set_extensions (NemoAction *action, gchar **extensions)
-{
-    gchar **tmp;
-
-    tmp = action->extensions;
-    action->extensions = g_strdupv (extensions);
-    g_strfreev (tmp);
-}
-
-static gchar **
-nemo_action_get_extension_list (NemoAction *action)
-{
-    return action->extensions;
-}
-
-static void
-nemo_action_set_mimetypes (NemoAction *action, gchar **mimetypes)
-{
-    gchar **tmp;
-
-    tmp = action->mimetypes;
-    action->mimetypes = g_strdupv (mimetypes);
-    g_strfreev (tmp);
-}
-
-static gchar **
-nemo_action_get_mimetypes_list (NemoAction *action)
-{
-    return action->mimetypes;
-}
-
-static void
-nemo_action_set_key_file_path (NemoAction *action, const gchar *path)
-{
-    gchar *tmp;
-    tmp = action->key_file_path;
-    action->key_file_path = g_strdup (path);
-    g_free (tmp);
-}
-
-static void
-nemo_action_set_exec (NemoAction *action, const gchar *exec)
-{
-    gchar *tmp;
-
-    tmp = action->exec;
-    action->exec = g_strdup (exec);
-    g_free (tmp);
-}
-
-static void
-nemo_action_set_parent_dir (NemoAction *action, const gchar *parent_dir)
-{
-    gchar *tmp;
-
-    tmp = action->parent_dir;
-    action->parent_dir = g_strdup (parent_dir);
-    g_free (tmp);
-}
-
-static void
-nemo_action_set_separator (NemoAction *action, const gchar *separator)
-{
-    gchar *tmp;
-
-    tmp = action->separator;
-    action->separator = g_strdup (separator);
-    g_free (tmp);
-}
-
-static void
-nemo_action_set_conditions (NemoAction *action, gchar **conditions)
-{
-    gchar **tmp;
-
-    tmp = action->conditions;
-    action->conditions = g_strdupv (conditions);
-    g_strfreev (tmp);
-}
-
-static gchar **
-nemo_action_get_conditions (NemoAction *action)
-{
-    return action->conditions;
-}
-
-static void
-nemo_action_set_orig_label (NemoAction *action, const gchar *orig_label)
-{
-    gchar *tmp;
-
-    tmp = action->orig_label;
-    action->orig_label = g_strdup (orig_label);
-    g_free (tmp);
-}
-
-static void
-nemo_action_set_orig_tt (NemoAction *action, const gchar *orig_tt)
-{
-    gchar *tmp;
-
-    tmp = action->orig_tt;
-    action->orig_tt = g_strdup (orig_tt);
-    g_free (tmp);
-}
-
 const gchar *
 nemo_action_get_orig_label (NemoAction *action)
 {
@@ -1677,18 +1610,6 @@ nemo_action_get_tt (NemoAction *action,
     return ret;
 }
 
-static gboolean
-get_dbus_satisfied (NemoAction *action)
-{
-    return action->dbus_satisfied;
-}
-
-static gboolean
-get_gsettings_satisfied (NemoAction *action)
-{
-    return action->gsettings_satisfied;
-}
-
 static gboolean
 check_exec_condition (NemoAction  *action,
                       const gchar *condition,
@@ -1783,19 +1704,21 @@ nemo_action_get_visibility (NemoAction *action,
                             GtkWindow  *window)
 {
     // Check DBUS
-    if (!get_dbus_satisfied (action))
+    if (!action->dbus_satisfied)
         return FALSE;
 
-    if (!get_gsettings_satisfied (action))
+    if (!action->gsettings_satisfied)
         return FALSE;
 
+    if ((action->uri_scheme != NULL) && !nemo_file_has_uri_scheme (parent, action->uri_scheme)) {
+        return FALSE;
+    }
+
     // Check selection
     gboolean selection_type_show = FALSE;
-    SelectionType selection_type = nemo_action_get_selection_type (action);
-
     guint selected_count = g_list_length (selection);
 
-    switch (selection_type) {
+    switch (action->selection_type) {
         case SELECTION_SINGLE:
             selection_type_show = selected_count == 1;
             break;
@@ -1812,7 +1735,7 @@ nemo_action_get_visibility (NemoAction *action,
             selection_type_show = TRUE;
             break;
         default:
-            selection_type_show = selected_count == selection_type;
+            selection_type_show = selected_count == action->selection_type;
             break;
     }
 
@@ -1821,8 +1744,8 @@ nemo_action_get_visibility (NemoAction *action,
 
     // Check extensions and mimetypes
     gboolean extension_type_show = TRUE;
-    gchar **extensions = nemo_action_get_extension_list (action);
-    gchar **mimetypes = nemo_action_get_mimetypes_list (action);
+    gchar **extensions = action->extensions;
+    gchar **mimetypes = action->mimetypes;
 
     guint ext_count = extensions != NULL ? g_strv_length (extensions) : 0;
     guint mime_count = mimetypes != NULL ? g_strv_length (mimetypes) : 0;
@@ -1897,7 +1820,7 @@ nemo_action_get_visibility (NemoAction *action,
 
     // Check conditions
     gboolean condition_type_show = TRUE;
-    gchar **conditions = nemo_action_get_conditions (action);
+    gchar **conditions = action->conditions;
     guint condition_count = conditions != NULL ? g_strv_length (conditions) : 0;
 
     if (condition_count > 0) {
