@@ -83,7 +83,9 @@ enum {
     LIST_VIEW,
     COMPACT_VIEW,
     SIDEBAR_PLACES,
-    SIDEBAR_TREE
+    SIDEBAR_TREE,
+    TOOLBAR_PATHBAR,
+    TOOLBAR_ENTRY
 };
 
 static void
@@ -145,7 +147,7 @@ action_home_callback (GtkAction *action,
         NemoWindowSlot *slot;
 
         window = NEMO_WINDOW (user_data);
-g_printerr ("what\n");
+
         slot = nemo_window_get_active_slot (window);
 
         nemo_window_slot_go_home (slot, nemo_event_get_window_open_flags ());
@@ -307,9 +309,11 @@ action_plugins_callback (GtkAction *action,
 {
     GtkWindow *window;
 
-    window = GTK_WINDOW (user_data);
+    if (!NEMO_IS_DESKTOP_WINDOW (user_data)) {
+        window = GTK_WINDOW (user_data);
 
-    nemo_file_management_properties_dialog_show (window, "plugins");
+        nemo_file_management_properties_dialog_show (window, "plugins");
+    }
 }
 
 static void
@@ -525,7 +529,9 @@ static void
 action_close_all_windows_callback (GtkAction *action,
 				   gpointer user_data)
 {
-	nemo_application_close_all_windows (nemo_application_get_singleton ());
+	if (!NEMO_IS_DESKTOP_WINDOW (user_data)) {
+		nemo_application_close_all_windows (nemo_application_get_singleton ());
+	}
 }
 
 static void
@@ -579,12 +585,14 @@ action_show_hide_sidebar_callback (GtkAction *action,
 {
 	NemoWindow *window;
 
-	window = NEMO_WINDOW (user_data);
+	if (!NEMO_IS_DESKTOP_WINDOW (user_data)) {
+		window = NEMO_WINDOW (user_data);
 
-	if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action))) {
-		nemo_window_show_sidebar (window);
-	} else {
-		nemo_window_hide_sidebar (window);
+		if (gtk_toggle_action_get_active (GTK_TOGGLE_ACTION (action))) {
+			nemo_window_show_sidebar (window);
+		} else {
+			nemo_window_hide_sidebar (window);
+		}
 	}
 }
 
@@ -719,6 +727,39 @@ view_radio_entry_changed_cb (GtkAction *action,
     }
 }
 
+static void
+toolbar_radio_entry_changed_cb (GtkAction *action,
+                                GtkRadioAction *current,
+                                gpointer user_data)
+{
+    NemoWindow *window = NEMO_WINDOW (user_data);
+    NemoWindowPane *pane;
+    GtkAction *toggle_action;
+    gint current_value;
+
+    if (NEMO_IS_DESKTOP_WINDOW (window)) {
+        return;
+    }
+
+    pane = nemo_window_get_active_pane (window);
+    toggle_action = gtk_action_group_get_action (pane->action_group, NEMO_ACTION_TOGGLE_LOCATION);
+
+    current_value = gtk_radio_action_get_current_value (current);
+    switch (current_value) {
+        case TOOLBAR_PATHBAR:
+            g_settings_set_boolean (nemo_preferences, NEMO_PREFERENCES_SHOW_LOCATION_ENTRY, FALSE);
+            gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (toggle_action), FALSE);
+            break;
+        case TOOLBAR_ENTRY:
+            g_settings_set_boolean (nemo_preferences, NEMO_PREFERENCES_SHOW_LOCATION_ENTRY, TRUE);
+            gtk_toggle_action_set_active (GTK_TOGGLE_ACTION (toggle_action), TRUE);
+            break;
+        default:
+            ;
+            break;
+    }
+}
+
 /* TODO: bind all of this with g_settings_bind and GBinding */
 static guint
 sidebar_id_to_value (const gchar *sidebar_id)
@@ -770,14 +811,18 @@ static void
 action_add_bookmark_callback (GtkAction *action,
 			      gpointer user_data)
 {
+    if (!NEMO_IS_DESKTOP_WINDOW (user_data)) {
         nemo_window_add_bookmark_for_current_location (NEMO_WINDOW (user_data));
+    }
 }
 
 static void
 action_edit_bookmarks_callback (GtkAction *action,
 				gpointer user_data)
 {
+    if (!NEMO_IS_DESKTOP_WINDOW (user_data)) {
         nemo_window_edit_bookmarks (NEMO_WINDOW (user_data));
+    }
 }
 
 static void
@@ -833,6 +878,24 @@ action_new_window_callback (GtkAction *action,
         GFile *loc;
 
         uri = nemo_window_slot_get_current_uri (nemo_window_get_active_slot (current_window));
+
+        if (eel_uri_is_search (uri)) {
+            NemoDirectory *dir;
+            NemoQuery *query;
+
+            dir = nemo_directory_get_by_uri (uri);
+            query = nemo_search_directory_get_query (NEMO_SEARCH_DIRECTORY (dir));
+
+            if (query != NULL) {
+                g_free (uri);
+
+                uri = nemo_query_get_location (query);
+                g_object_unref (query);
+            }
+
+            nemo_directory_unref (dir);
+        }
+
         loc = g_file_new_for_uri (uri);
 
         application = nemo_application_get_singleton ();
@@ -860,15 +923,14 @@ action_new_tab_callback (GtkAction *action,
 void action_toggle_location_entry_callback (GtkToggleAction *action, gpointer user_data);
 
 static void
-toggle_location_entry_setting (NemoWindow     *window,
-                               NemoWindowPane *pane,
-                               gboolean        from_accel_or_menu)
+toggle_location_entry (NemoWindow     *window,
+                       NemoWindowPane *pane,
+                       gboolean        from_accel_or_menu)
 {
     gboolean current_view, temp_toolbar_visible, default_toolbar_visible, grab_focus_only, already_has_focus;
     GtkToggleAction *button_action;
     GtkActionGroup *action_group;
 
-
     current_view = nemo_toolbar_get_show_location_entry (NEMO_TOOLBAR (pane->tool_bar));
     temp_toolbar_visible = pane->temporary_navigation_bar;
     default_toolbar_visible = g_settings_get_boolean (nemo_window_state,
@@ -879,7 +941,6 @@ toggle_location_entry_setting (NemoWindow     *window,
 
     if ((temp_toolbar_visible || default_toolbar_visible) && !grab_focus_only) {
         nemo_toolbar_set_show_location_entry (NEMO_TOOLBAR (pane->tool_bar), !current_view);
-        g_settings_set_boolean (nemo_preferences, NEMO_PREFERENCES_SHOW_LOCATION_ENTRY, !current_view);
 
         action_group = pane->toolbar_action_group;
         button_action = GTK_TOGGLE_ACTION (gtk_action_group_get_action (action_group, NEMO_ACTION_TOGGLE_LOCATION));
@@ -900,14 +961,14 @@ action_toggle_location_entry_callback (GtkToggleAction *action,
     NemoWindowPane *pane;
 
     pane = nemo_window_get_active_pane (window);
-    toggle_location_entry_setting(window, pane, FALSE);
+    toggle_location_entry (window, pane, FALSE);
 }
 
 void nemo_window_show_location_entry (NemoWindow *window) {
 	NemoWindowPane *pane;
 
     pane = nemo_window_get_active_pane (window);
-    toggle_location_entry_setting(window, pane, TRUE);
+    toggle_location_entry (window, pane, TRUE);
 }
 
 static void
@@ -917,8 +978,10 @@ action_menu_edit_location_callback (GtkAction *action,
 	NemoWindow *window = user_data;
 	NemoWindowPane *pane;
 
-    pane = nemo_window_get_active_pane (window);
-    toggle_location_entry_setting(window, pane, TRUE);
+    if (!NEMO_IS_DESKTOP_WINDOW (user_data)) {
+        pane = nemo_window_get_active_pane (window);
+        toggle_location_entry (window, pane, TRUE);
+    }
 }
 
 static void
@@ -1463,7 +1526,8 @@ static const GtkActionEntry main_entries[] = {
   { "TabsMoveRight", NULL, N_("Move Tab _Right"), "<shift><control>Page_Down",
     N_("Move current tab to right"),
     G_CALLBACK (action_tabs_move_right_callback) },
-  { "Sidebar List", NULL, N_("Sidebar") }
+  { "Sidebar List", NULL, N_("Sidebar") },
+  { "Toolbar List", NULL, N_("Toolbar") }
 };
 
 static const GtkToggleActionEntry main_toggle_entries[] = {
@@ -1530,6 +1594,15 @@ static const GtkRadioActionEntry view_radio_entries[] = {
       COMPACT_VIEW }
 };
 
+static const GtkRadioActionEntry toolbar_radio_entries[] = {
+    { NEMO_ACTION_TOOLBAR_ALWAYS_SHOW_PATHBAR, NULL,
+      N_("Path Bar"), NULL, N_("Always prefer the path bar"),
+      TOOLBAR_PATHBAR },
+    { NEMO_ACTION_TOOLBAR_ALWAYS_SHOW_ENTRY, NULL,
+      N_("Location Entry"), NULL, N_("Always prefer the location entry"),
+      TOOLBAR_ENTRY }
+};
+
 GtkActionGroup *
 nemo_window_create_toolbar_action_group (NemoWindow *window)
 {
@@ -1860,6 +1933,14 @@ nemo_window_initialize_menus (NemoWindow *window)
                         view_radio_entries, G_N_ELEMENTS (view_radio_entries),
                         0, G_CALLBACK (view_radio_entry_changed_cb),
                         window);
+
+    gboolean use_entry = g_settings_get_boolean (nemo_preferences, NEMO_PREFERENCES_SHOW_LOCATION_ENTRY);
+    gtk_action_group_add_radio_actions (action_group,
+                                        toolbar_radio_entries, G_N_ELEMENTS (toolbar_radio_entries),
+                                        use_entry ? TOOLBAR_ENTRY : TOOLBAR_PATHBAR,
+                                        G_CALLBACK (toolbar_radio_entry_changed_cb),
+                                        window);
+
 	action = gtk_action_group_get_action (action_group, NEMO_ACTION_UP);
 	g_object_set (action, "short_label", _("_Up"), NULL);
 
